----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    17:20:35 09/12/2010 
-- Design Name: 
-- Module Name:    fifo_fwft_001 - arch 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
-- FWFT This is a simple First Word Fall Through FWFT circular FIFO
-- based on the FIFO design presented in P CHU's book
-- "FPGA Programming by VHDL examples"
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity fifo_fwft_001 is

	generic	(
				Data_Bits: natural:=8;  	-- number of data bits 
				Address_Bits: natural:=3	-- number of address bits
				);

   Port    ( 
			  clk 				 	:	in  std_logic;
           reset		 		 	: 	in  std_logic;
           i_fifo_rd			  	: 	in  std_logic;
           i_fifo_wr			 	: 	in  std_logic;
           i_fifo_wr_data 		: 	in  std_logic_vector (Data_Bits-1 downto 0);
			  
           o_fifo_empty 		: 	out  std_logic;
           o_fifo_full 			: 	out  std_logic;
           o_fifo_rd_data 		: 	out  std_logic_vector (Data_Bits-1 downto 0)
			  );

end fifo_fwft_001;

architecture arch of fifo_fwft_001 is

	type register_type is array ((2**Address_Bits)-1 downto 0) of 
		std_logic_vector(Data_Bits-1 downto 0);
		
	signal t_array_reg			: register_type;
	signal slv_write_ptr_now 	: std_logic_vector(Address_Bits-1 downto 0);
	signal slv_write_ptr_next 	: std_logic_vector(Address_Bits-1 downto 0);
	signal slv_write_ptr_inc	: std_logic_vector(Address_Bits-1 downto 0);
	
	signal slv_read_ptr_now		: std_logic_vector(Address_Bits-1 downto 0);
	signal slv_read_ptr_next	: std_logic_vector(Address_Bits-1 downto 0);
	signal slv_read_ptr_inc		: std_logic_vector(Address_Bits-1 downto 0);
	
	signal slv_read_data_now	: std_logic_vector(Data_Bits-1 downto 0);
	signal slv_read_data_next	: std_logic_vector(Data_Bits-1 downto 0);
	
	signal slv_write_data		: std_logic_vector(Data_Bits-1 downto 0);
	
	signal slv_fifo_operator	: std_logic_vector(1 downto 0);
	
	signal sl_fifo_full_now		: std_logic;
	signal sl_fifo_full_next	: std_logic;
	signal sl_fifo_empty_now	: std_logic;
	signal sl_fifo_empty_next	: std_logic;
	
	signal sl_write_enable		: std_logic;

begin

	-- Clocking Process
	process(clk,reset)
	begin
		if (reset='1') then
			t_array_reg <= (others =>(others=>'0'));
			slv_write_ptr_now <= (others => '0');
			slv_read_ptr_now <= (others => '0');
			sl_fifo_full_now <= '0';
			sl_fifo_empty_now <= '1';
		elsif (clk'event and clk='1') then
			
			slv_read_ptr_now <= slv_read_ptr_next;
			slv_read_data_now <= slv_read_data_next;
			sl_fifo_empty_now <= sl_fifo_empty_next;
			sl_fifo_full_now <= sl_fifo_full_next;
			slv_write_ptr_now <= slv_write_ptr_next;
			if (sl_write_enable='1') then
				t_array_reg(to_integer(unsigned(slv_write_ptr_now))) <= slv_write_data;				
			end if;		
		end if;
	end process;
	
	-- I/O ASSIGNMENTS
	--Outputs
	o_fifo_rd_data <= slv_read_data_now;
	o_fifo_full <= sl_fifo_full_now;
	o_fifo_empty <= sl_fifo_empty_now;
	-- Inputs
	slv_write_data <= i_fifo_wr_data;
	
	-- OTHER ASSIGNMENTS
	sl_write_enable <= i_fifo_wr and (not sl_fifo_full_now);
	
	-- This computes the next pointer values
	slv_write_ptr_inc <= std_logic_vector(unsigned(slv_write_ptr_now)+1);
	slv_read_ptr_inc <= std_logic_vector(unsigned(slv_read_ptr_now)+1);
	
	slv_fifo_operator <= i_fifo_wr & i_fifo_rd;
	
	-- PROCESSES
	
	-- Data Output
	-- Note that the outputs are forced to 0 if the fifo register is empty!
	process(sl_fifo_empty_next, slv_read_ptr_next, t_array_reg, sl_fifo_empty_now,
		i_fifo_wr_data)
	begin
		slv_read_data_next  <= t_array_reg(to_integer(unsigned(slv_read_ptr_next)));
		
		if (sl_fifo_empty_next = '1') then
			-- if fifo is empty output '0'
			-- (not really necessary but is useful during simulation)
			slv_read_data_next  <= (others => '0');
		else
			if (sl_fifo_empty_now = '1') then
				-- If fifo empty next = 0 and fifo empty now = 1 then 
				-- feed data straight through to output on this clock cycle
				-- (First Word Fall Through)
				slv_read_data_next <= i_fifo_wr_data;	
			end if;
		end if;	
	end process;
	
	process (slv_write_ptr_now, slv_write_ptr_inc, slv_read_ptr_now, slv_read_ptr_inc, 
		slv_fifo_operator, sl_fifo_empty_now, sl_fifo_full_now)
	
	begin
		slv_write_ptr_next <= slv_write_ptr_now;
		slv_read_ptr_next <= slv_read_ptr_now;
		sl_fifo_full_next <= sl_fifo_full_now;
		sl_fifo_empty_next <= sl_fifo_empty_now;
		
		case slv_fifo_operator is
			when "00" => -- This means no operation
			
			when "01" => --read
				if (sl_fifo_empty_now /= '1') then --not empty
					slv_read_ptr_next <= slv_read_ptr_inc;
					sl_fifo_full_next <= '0';
					if (slv_read_ptr_inc=slv_write_ptr_now) then
						sl_fifo_empty_next <= '1';
					end if;
				end if;
			
			when "10" => -- write
				if (sl_fifo_full_now /= '1') then -- not full
					slv_write_ptr_next <= slv_write_ptr_inc;
					sl_fifo_empty_next <= '0';
					if (slv_write_ptr_inc=slv_read_ptr_now) then
						sl_fifo_full_next <= '1';
					end if;
				end if;
				
			when others => --read/write
				slv_write_ptr_next <= slv_write_ptr_inc;
				slv_read_ptr_next <= slv_read_ptr_inc;
			end case;
		end process;

end arch;

