-----------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
-----------------------------------------------------------


-- Entity
-----------------------------------------------------------
entity trigger_delay is
	generic(
		G_counter_width : natural := 15
	);
	port(
		I_rst			: in	std_logic;
		I_clk			: in	std_logic;
		--
		I_delay		: in	std_logic_vector(G_counter_width - 1 downto 0);
		I_trig		: in	std_logic;
		--
		O_trig_out	: out	std_logic;
		--
		O_ready		: out std_logic
	);
end entity trigger_delay;

-- Architecture
-----------------------------------------------------------

-- v1: The previous version had too much latency (due to the
--     state machine). Also, it contained a bug which meant that
--     the output triggers were 2 clks long. Note that another output
--     has also been added: 0_ready
architecture v1 of trigger_delay is
	
	-- Values / Signals
	-----------------------------
	
	constant CA_min_counter_value : std_logic_vector(G_counter_width - 1 downto 0) := (others => '0');
	
	signal SA_delay_counter : std_logic_vector(G_counter_width - 1 downto 0);
	signal SA_ready			: std_logic;
	
begin
	
	-- Processes
	-----------------------------
	
	delay_trigger: process(I_rst, I_clk)
	begin
		
		if(I_rst = '1') then
			
			SA_delay_counter <= (others => '0');
			SA_ready <= '1';
			
			O_trig_out <= '0';
			
		elsif(I_clk'event and I_clk = '1') then
			
			O_trig_out <= '0';
			
			if(SA_ready = '1') then
				
				SA_delay_counter <= I_delay;
				
				if(I_trig = '1') then
					SA_ready <= '0';
				end if;
				
			else
				
				if(SA_delay_counter = CA_min_counter_value) then
					O_trig_out <= '1';
					SA_ready <= '1';
				else
					SA_delay_counter <= SA_delay_counter - '1';
				end if;
				
			end if;
			
		end if;
		
	end process delay_trigger;
	
	-- Concurrent signal
	-- assignment
	-----------------------------
	
	O_ready <= SA_ready;
	
end architecture v1;


-- v0: First attempt
--architecture v0 of trigger_delay is
--	
--	-- Types
--	-----------------------------
--	
--	-- This isn't really necessary for such a simple system,
--	-- but it makes the code slightly tidier...
--	type TA_delay_state is
--	(
--		idle,
--		counting
--	);
--	
--	-- Values / Signals
--	-----------------------------
--	
--	constant CA_min_counter_value : std_logic_vector(G_counter_width - 1 downto 0) := (others => '0');
--	
--	signal SA_state			: TA_delay_state;
--	signal SA_delay_counter	: std_logic_vector(G_counter_width - 1 downto 0);
--	
--begin
--	
--	-- Processes
--	-----------------------------
--	
--	delay_trigger: process(I_rst, I_clk)
--	begin
--		
--		if(I_rst = '1') then
--			
--			SA_delay_counter <= (others => '0');
--			O_trig_out <= '0';
--			SA_state <= idle;
--			
--		elsif(I_clk'event and I_clk = '1') then
--			
--			case SA_state is
--				
--				when idle =>
--					
--					SA_delay_counter <= I_delay;
--					O_trig_out <= '0';
--					
--					if(I_trig = '1') then
--						SA_state <= counting;
--					end if;
--				
--				when counting =>
--					
--					if(SA_delay_counter = CA_min_counter_value) then
--						O_trig_out <= '1';
--						SA_state <= idle;
--					else
--						SA_delay_counter <= SA_delay_counter - '1';
--					end if;
--				
--				-- Cannot happen, but include for completeness...
--				when others =>
--					SA_state <= idle;
--				
--			end case;
--			
--		end if;
--		
--	end process delay_trigger;
--	
--end architecture v0;
